import torch
from .context_managers import temp_mode_change
from contextlib import contextmanager



def get_paramnames_with_no_gradient(model):
    return [name for name, param in model.named_parameters() if param.grad is None and param.requires_grad]


def get_output_shape_of_model(model, forward_fn, **forward_kwargs):
    with temp_mode_change(model=model, mode=False):
        # get outputshape from forward pass
        x = torch.ones(1, *model.input_shape, device=model.device)
        output = forward_fn(x, **forward_kwargs)
    return tuple(output.shape[1:])



@torch.no_grad()
def copy_params(source_model, target_model):
    # TODO i think inplace operations are okay here
    for target_param, source_param in zip(target_model.parameters(), source_model.parameters()):
        target_param.data = source_param.data


@torch.no_grad()
def update_ema(source_model, target_model, target_factor):
    # TODO i think inplace operations are okay here
    for target_param, source_param in zip(target_model.parameters(), source_model.parameters()):
        target_param.data = target_param.data * target_factor + source_param.data * (1. - target_factor)
    for target_buffer, source_buffer in zip(target_model.buffers(), source_model.buffers()):
        target_buffer.data.copy_(source_buffer.data)


def get_trainable_param_count(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def get_frozen_param_count(model):
    return sum(p.numel() for p in model.parameters() if not p.requires_grad)


@torch.no_grad()
def backup_all_buffers(model):
    buffers = {}
    for name, buffer in model.named_buffers():
        buffers[name] = buffer.clone()
    return buffers

@torch.no_grad()
def restore_all_buffers(model, buffers):
    for name, buffer in model.named_buffers():
        buffer.data.copy_(buffers[name])
    return buffers

@contextmanager
def preserve_buffers(model):
    buffer_bkp = backup_all_buffers(model=model)
    yield
    restore_all_buffers(model=model, buffers=buffer_bkp)





from contextlib import contextmanager


MODEL_MODES = {
    "train": True,
    "training": True,
    "eval": False,
    "evaluate": False,
}


@contextmanager
def temp_mode_change(model, mode):
    if mode in MODEL_MODES:
        mode = MODEL_MODES[mode]
    assert isinstance(mode, bool)

    prev_mode = model.training
    model.train(mode=mode)
    yield
    if prev_mode != model.training:
        model.train(mode=prev_mode)




import gc

import torch


def get_tensors_in_memory():
    # some warning was thrown when calling torch.is_tensor(_reduce_op) with a _reduce_op object
    all_objs = gc.get_objects()
    all_tensors = []
    cuda_tensors = []
    for obj in all_objs:
        try:
            if type(obj).__name__ != "_reduce_op" and torch.is_tensor(obj):
                all_tensors.append(obj)
                if obj.device != torch.device("cpu"):
                    cuda_tensors.append(obj)
        except ReferenceError:
            # with wandb there is some issue where 'ReferenceError: weakly-referenced object no longer exists' is raised
            pass
    return all_tensors, cuda_tensors



import torch
from torch.cuda.amp import GradScaler

FLOAT32_ALIASES = ["float32", 32]
FLOAT16_ALIASES = ["float16", 16]
BFLOAT16_ALIASES = ["bfloat16", "bf16"]
VALID_PRECISIONS = FLOAT32_ALIASES + FLOAT16_ALIASES + BFLOAT16_ALIASES


def get_supported_precision(desired_precision, device, log_fn=print):
    assert desired_precision in VALID_PRECISIONS
    if desired_precision in FLOAT32_ALIASES:
        return torch.float32
    if desired_precision in FLOAT16_ALIASES:
        desired_precision = "float16"
    if desired_precision in BFLOAT16_ALIASES:
        desired_precision = "bfloat16"

    if desired_precision == "bfloat16":
        if is_bfloat16_compatible(device):
            return torch.bfloat16
        else:
            # old cuda devices don't support bfloat16
            if is_float16_compatible(device):
                log_fn("bfloat16 not supported -> using float32 (float16 could lead to under-/overflows)")
                return torch.float32

    if desired_precision == "float16":
        if is_float16_compatible(device):
            return torch.float16
        else:
            # currently cpu only supports bfloat16
            if is_bfloat16_compatible(device):
                log_fn(f"float16 not supported -> using bfloat16")
                return torch.bfloat16

    log_fn(f"float16/bfloat16 not supported -> using float32")
    return torch.float32


def _is_compatible(device, dtype):
    try:
        with torch.autocast(device_type=str(device), dtype=dtype):
            pass
    except RuntimeError:
        return False
    return True


def is_bfloat16_compatible(device):
    return _is_compatible(device, torch.bfloat16)


def is_float16_compatible(device):
    return _is_compatible(device, torch.float16)


class NoopGradScaler:
    @staticmethod
    def scale(loss):
        return loss

    @staticmethod
    def unscale_(optimizer):
        pass

    @staticmethod
    def step(optimizer, *args, **kwargs):
        optimizer.step(*args, **kwargs)

    @staticmethod
    def update():
        pass

class NoopContext:
    def __enter__(self):
        pass

    def __exit__(self, *args, **kwargs):
        pass


def get_grad_scaler_and_autocast_context(precision, device):
    if precision == torch.float32:
        return NoopGradScaler(), NoopContext()
    if precision == torch.bfloat16:
        # GradScaler shouldn't be necessary (https://github.com/pytorch/pytorch/issues/36169)
        return NoopGradScaler(), torch.autocast(str(device), dtype=precision)
    elif precision == torch.float16:
        return GradScaler(), torch.autocast(str(device), dtype=precision)
    raise NotImplementedError



import einops
import numpy as np
import png
import torch
from PIL import Image
from torchvision.transforms.functional import to_tensor

from .constants import VIRIDIS_PALETTE_NP


def png_loader(path):
    with open(path, "rb") as f:
        return to_tensor(Image.open(f))


def png_loader_viridis(path):
    img = (png_loader(path) * 255).squeeze(0).long().numpy()
    rgb = np.take(VIRIDIS_PALETTE_NP, img, axis=0)
    return einops.rearrange(torch.from_numpy(rgb), "h w c -> c h w")


def png_loader_with_info(path):
    # might also be possible with Image.open (is possibly faster)
    r = png.Reader(filename=str(path))
    w, h, reader, info = r.read()
    data = []
    for row in reader:
        data.append(torch.frombuffer(row, dtype=torch.int8))
    data = torch.stack(data)
    return data.unsqueeze(0) / 255, info



import collections.abc
from itertools import repeat
from pathlib import Path


# adapted from timm (timm/models/layers/helpers.py)
def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            assert len(x) == n
            return x
        return tuple(repeat(x, n))

    return parse


def _is_ntuple(n):
    def check(x):
        return isinstance(x, tuple) and len(param) == n

    return check


def to_ntuple(x, n):
    return _ntuple(n=n)(x)


def is_ntuple(x, n):
    return _is_ntuple(n=n)(x)


to_2tuple = _ntuple(2)
is_2tuple = _is_ntuple(2)
to_3tuple = _ntuple(3)
is_3tuple = _is_ntuple(3)
to_4tuple = _ntuple(4)
is_4tuple = _is_ntuple(4)
to_5tuple = _ntuple(5)
is_5tuple = _is_ntuple(5)


def float_to_integer_exact(f):
    assert f.is_integer()
    return int(f)


def check_exclusive(*args):
    return sum(arg is not None for arg in args) == 1


def check_inclusive(*args):
    return sum(arg is not None for arg in args) in [0, len(args)]


def check_at_least_one(*args):
    return sum(arg is not None for arg in args) > 0


def check_at_most_one(*args):
    return sum(arg is not None for arg in args) <= 1


def to_path(path):
    if path is not None and not isinstance(path, Path):
        return Path(path).expanduser()
    return path


def to_list_of_values(list_or_item, default_value=None):
    if list_or_item is None:
        if default_value is None:
            return []
        else:
            return [default_value]
    if not isinstance(list_or_item, (tuple, list)):
        return [list_or_item]
    return list_or_item



from torchvision.transforms.functional import to_pil_image

from .constants import VIRIDIS_PALETTE_LIST


def png_writer_greyscale(img, fp):
    if img.ndim == 2:
        img = img.unsqueeze(0)
    assert img.ndim == 3 and img.size(0) == 1
    to_pil_image(img, mode="L").save(fp)


def png_writer_viridis(img, fp):
    if img.ndim == 2:
        img = img.unsqueeze(0)
    assert img.ndim == 3 and img.size(0) == 1
    img = to_pil_image(img, mode="L")
    img.putpalette(VIRIDIS_PALETTE_LIST)
    img.save(fp)



import torch
from kappautils.param_checking import to_2tuple

def xy_to_image(x, y, resolution, xmin=None, xmax=None, ymin=None, ymax=None, weights=None):
    coords = torch.stack([x, y], dim=1)
    if xmin is not None:
        assert xmax is not None
        coords_min = torch.tensor(xmin, xmax, device=x.device)
    else:
        coords_min = None
    if ymin is not None:
        assert ymax is not None
        coords_max = torch.tensor(xmin, xmax, device=x.device)
    else:
        coords_max = None
    return coords_to_image(
        coords=coords,
        resolution=resolution,
        coords_min=coords_min,
        coords_max=coords_max,
        weights=weights,
    )


def coords_to_image(coords, resolution, coords_min=None, coords_max=None, weights=None):
    if not torch.is_tensor(resolution):
        resolution = torch.tensor(to_2tuple(resolution), device=coords.device)
    assert resolution.ndim == 1 and resolution.numel() == 2
    # rescale coords to resolution
    if coords_min is not None:
        if not torch.is_tensor(coords_min):
            coords_min = torch.tensor(coords_min, device=coords.device)
    else:
        coords_min = coords.min(dim=0).values
    assert coords_min.ndim == 1 and coords_min.numel() == 2
    if coords_max is not None:
        if not torch.is_tensor(coords_max):
            coords_max = torch.tensor(coords_max, device=coords.device)
    else:
        coords_max = coords.max(dim=0).values
    assert coords_max.ndim == 1 and coords_max.numel() == 2
    coords_max = coords_max / (resolution - 1)
    coords = (coords - coords_min.unsqueeze(0)) / coords_max.unsqueeze(0)
    coords = coords.round().long()
    # flatten coords (e.g. [4, 6] with resolution=(6, 8) will be flattened to 4 * 8 + 6)
    coords = coords[:, 0] * resolution[1] + coords[:, 1]
    # weighted counting of occourances
    img = torch.bincount(coords, weights=weights, minlength=resolution[0] * resolution[1])
    img = img.view(*resolution)
    return img
